# flake8: noqa
from __future__ import annotations

import bz2
import hashlib
import logging
import os
import re

import requests

from vunnel import utils
from vunnel.utils.oval_parser import Config, parse


class Parser:
    _url_mappings_ = [
        # Legacy data for RHEL:5 - no longer available from endpoint after 1st July, 2023; however, it is available in the
        # preload archive.
        {
            "base_url": "https://www.redhat.com/security/data/oval",
            "manifest_path": "PULP_MANIFEST",
            "oval_paths": ["com.redhat.rhsa-all.xml.bz2"],
            "skip_download": True,
        },
        {
            "base_url": "https://www.redhat.com/security/data/oval/v2",
            "manifest_path": "PULP_MANIFEST",
            "oval_paths": [
                "RHEL6/rhel-6.oval.xml.bz2",
                "RHEL7/rhel-7.oval.xml.bz2",
                "RHEL8/rhel-8.oval.xml.bz2",
                "RHEL9/rhel-9.oval.xml.bz2",
            ],
        },
    ]

    def __init__(self, workspace, config: Config, logger=None, download_timeout=125):
        self.config = config
        self.download_timeout = download_timeout
        self.workspace = workspace

        if not logger:
            logger = logging.getLogger(self.__class__.__name__)

        self.logger = logger
        self._urls = set()

    @property
    def urls(self) -> list[str]:
        return list(self._urls)

    def _get_workspace_xml_filepath(self, oval_url_path: str) -> str:
        return os.path.join(self.workspace.input_path, oval_url_path.rstrip(".bz2"))

    def _get_sha256_map(self, base_url: str, manifest_path: str, oval_paths: list[str]) -> dict[str, str]:
        try:
            manifest_url = f"{base_url}/{manifest_path}"
            self.logger.info(f"begin processing manifest from {manifest_url}")
            path_to_sha = {}
            r = requests.get(manifest_url, timeout=self.download_timeout)
            if r.status_code == 200:
                unmatched_oval_paths = set(oval_paths.copy())

                for line in r.text.splitlines():
                    matched_path = None
                    for p in unmatched_oval_paths:
                        pattern = re.compile(rf"{p},([^,]+),.*")
                        match = pattern.match(line.strip())
                        if match:
                            path_to_sha[p] = match.group(1)
                            matched_path = p
                            break

                    if matched_path:
                        unmatched_oval_paths.remove(matched_path)

                    if len(unmatched_oval_paths) == 0:
                        break

                if len(unmatched_oval_paths) > 0:
                    error = f"no matching sha256 found for {unmatched_oval_paths}"
                    self.logger.error(error)
                    raise Exception(error)

                self.logger.info(f"finish processing manifest from {manifest_url}")
                return path_to_sha
            else:
                error = f"GET {manifest_url} failed with HTTP error {r.status_code}"
                self.logger.error(error)
                raise Exception(error)
        except:
            raise Exception("Error fetching/processing sha256")

    @utils.retry_with_backoff()
    def _download_oval_file(self, base_url: str, oval_url_path: str, path_to_sha: dict[str, str]) -> str:
        download = True
        xml_file_path = self._get_workspace_xml_filepath(oval_url_path=oval_url_path)
        xml_sha256_file_path = f"{xml_file_path}.sha256sum"
        previous_sha256 = None
        self.logger.info(f"comparing hashes for oval file {xml_file_path}")

        if os.path.exists(xml_file_path) and os.path.exists(xml_sha256_file_path):
            with open(xml_sha256_file_path) as fp:
                previous_sha256 = fp.read().strip()

        latest_sha256 = path_to_sha.get(oval_url_path)

        if previous_sha256 and latest_sha256:
            self.logger.debug(f"{oval_url_path}: previous sha256: {previous_sha256}, latest sha256: {latest_sha256}")
            download = latest_sha256.lower() != latest_sha256.lower()

        if download:
            try:
                oval_url = f"{base_url}/{oval_url_path}"
                self.logger.info(f"begin downloading OVAL file from {oval_url}")
                r = requests.get(oval_url, stream=True, timeout=self.download_timeout)
                if r.status_code == 200:
                    # compute the sha256 as the file is decompressed
                    sha256 = hashlib.sha256()
                    # ensure any nested directories get created
                    os.makedirs(os.path.dirname(xml_file_path), exist_ok=True)
                    with open(xml_file_path, "wb") as extracted:
                        decompressor = bz2.BZ2Decompressor()
                        for chunk in r.iter_content(chunk_size=1024):
                            uncchunk = decompressor.decompress(chunk)
                            extracted.write(uncchunk)
                            sha256.update(uncchunk)

                    self.logger.info(f"finish downloading OVAL file from {oval_url}")
                    sha256sum = str(sha256.hexdigest()).lower()
                    self.logger.debug(f"sha256 for {xml_file_path}: {sha256sum}")

                    # save the sha256 to another file
                    with open(xml_sha256_file_path, "w") as fp:
                        fp.write(sha256sum)

                    return sha256sum
                else:
                    raise Exception(f"GET {oval_url} failed with HTTP error {r.status_code}")
            except Exception:
                self.logger.exception("error downloading OVAL file")
                raise Exception("error downloading OVAL file")
        else:
            self.logger.info(f"stored checksum matches server checksum for {xml_file_path}. Skipping download")
            return latest_sha256

    @utils.retry_with_backoff()
    def _download(self):
        for m in self._url_mappings_:
            base_url = m["base_url"]
            manifest_path = m["manifest_path"]
            oval_paths = m["oval_paths"]
            skip_download = m.get("skip_download", False)
            path_to_sha = None

            if not skip_download:
                path_to_sha = self._get_sha256_map(base_url, manifest_path, oval_paths)

            self._urls.add(f"{base_url}/{manifest_path}")

            for p in oval_paths:
                self._urls.add(f"{base_url}/{p}")

                if not skip_download:
                    self._download_oval_file(base_url, p, path_to_sha)

        return None

    def xml_paths(self):
        paths = []
        for m in self._url_mappings_:
            oval_paths = m["oval_paths"]
            skip_download = m.get("skip_download", False)

            for p in oval_paths:
                # normalize and return results
                file_path = self._get_workspace_xml_filepath(p)
                if skip_download and not os.path.exists(file_path):
                    self.logger.warning(f"skip processing OVAL file {p}")
                    continue
                paths.append(file_path)
        return paths

    def parse(self):
        vuln_dict = {}

        for file_path in self.xml_paths():
            self.logger.info(f"begin parsing OVAL file {file_path}")
            partial_results = parse(file_path, self.config, vuln_dict=vuln_dict)
            vuln_dict.update(partial_results)
            self.logger.info(f"finish parsing OVAL file {file_path}")

        return vuln_dict

    def get(self):
        self._download()
        return self.parse()



from __future__ import annotations

import abc
import datetime
import enum
import logging
import time
from dataclasses import dataclass, field
from typing import Any

from . import result, workspace


class OnErrorAction(str, enum.Enum):
    FAIL = "fail"
    SKIP = "skip"
    RETRY = "retry"

    def __repr__(self) -> str:
        return self.value


class InputStatePolicy(str, enum.Enum):
    KEEP = "keep"
    DELETE = "delete"

    def __repr__(self) -> str:
        return self.value


class ResultStatePolicy(str, enum.Enum):
    KEEP = "keep"
    DELETE = "delete"
    DELETE_BEFORE_WRITE = "delete-before-write"  # treat like "KEEP" in error cases

    def __repr__(self) -> str:
        return self.value


@dataclass
class OnErrorConfig:
    # the action to take when an error occurs
    action: OnErrorAction = OnErrorAction.FAIL
    # the number of times to retry an action that fails
    retry_count: int = 3
    # the number of seconds to wait between retries
    retry_delay: int = 5
    # what to do with the input directory when an error occurs
    input: InputStatePolicy = InputStatePolicy.KEEP  # noqa: A003
    # what to do with the result directory when an error occurs
    results: ResultStatePolicy = ResultStatePolicy.KEEP

    def __post_init__(self) -> None:
        if not isinstance(self.action, OnErrorAction):
            self.action = OnErrorAction(self.action)
        if not isinstance(self.input, InputStatePolicy):
            self.input = InputStatePolicy(self.input)
        if not isinstance(self.results, ResultStatePolicy):
            self.results = ResultStatePolicy(self.results)


@dataclass
class RuntimeConfig:
    # what to do when an error occurs while the provider is running
    on_error: OnErrorConfig = field(default_factory=OnErrorConfig)
    # what to do with existing data in the input directory before running
    existing_input: InputStatePolicy = InputStatePolicy.KEEP
    # what to do with existing data in the result directory before running
    existing_results: ResultStatePolicy = ResultStatePolicy.KEEP
    # the format the results should be written in
    result_store: result.StoreStrategy = result.StoreStrategy.FLAT_FILE

    def __post_init__(self) -> None:
        if not isinstance(self.existing_input, InputStatePolicy):
            self.existing_input = InputStatePolicy(self.existing_input)
        if not isinstance(self.existing_results, ResultStatePolicy):
            self.existing_results = ResultStatePolicy(self.existing_results)
        if not isinstance(self.result_store, result.StoreStrategy):
            self.result_store = result.StoreStrategy(self.result_store)

    @property
    def skip_if_exists(self) -> bool:
        return self.existing_input == InputStatePolicy.KEEP


def disallow_existing_input_policy(cfg: RuntimeConfig) -> None:
    if cfg.existing_input != InputStatePolicy.KEEP:
        raise ValueError(
            f"existing_input policy is '{cfg.existing_input}' but only a value of 'keep' is allowed for this provider",
        )


class Provider(abc.ABC):
    # a breaking change to the semantics or values that the provider writes out should incur a version bump here.
    # this is used to determine if the provider can be run on an existing workspace or if it must be cleared first
    # (regardless of the existing_input and existing_result policy is).
    __version__: int = 1

    def __init__(self, root: str, runtime_cfg: RuntimeConfig = RuntimeConfig()):  # noqa: B008
        self.logger = logging.getLogger(self.name())
        self.workspace = workspace.Workspace(root, self.name(), logger=self.logger, create=False)
        self.urls: list[str] = []
        self.runtime_cfg = runtime_cfg

    @classmethod
    def version(cls) -> int:
        return cls.__version__

    @classmethod
    @abc.abstractmethod
    def name(cls) -> str:
        # note: wrapping [abstractmethod -> classmethod -> property] is no longer supported in
        # python 3.11 (deprecated due to undefined behavior) for this reason a simple classmethod
        # is used instead of a property
        raise NotImplementedError("'name()' must be implemented")

    @abc.abstractmethod
    def update(self, last_updated: datetime.datetime | None) -> tuple[list[str], int]:
        """Populates the input directory from external sources, processes the data, places results into the output directory."""
        raise NotImplementedError("'update()' must be implemented")

    def read_state(self) -> workspace.State | None:
        try:
            return workspace.State.read(root=self.workspace.path)
        except FileNotFoundError:
            return None

    def _update(self) -> None:
        start = datetime.datetime.now(tz=datetime.timezone.utc)

        last_updated = None
        current_state = self.read_state()
        if current_state:
            last_updated = current_state.timestamp

        urls, count = self.update(last_updated=last_updated)
        if count > 0:
            self.workspace.record_state(
                version=self.version(),
                timestamp=start,
                urls=urls,
                store=self.runtime_cfg.result_store.value,
            )
        else:
            self.logger.debug("skipping recording of workspace state (no new results found)")

    def run(self) -> None:
        self.logger.debug(f"using {self.workspace.path!r} as workspace")

        current_state = self.read_state()
        if current_state and current_state.version != self.version():
            self.logger.warning(f"provider version has changed from {current_state.version} to {self.version()}")
            self.logger.warning("clearing workspace to ensure consistency of existing input and results")
            self.workspace.clear()
        else:
            if self.runtime_cfg.existing_results == ResultStatePolicy.DELETE:
                self.workspace.clear_results()

            if self.runtime_cfg.existing_input == InputStatePolicy.DELETE:
                self.workspace.clear_input()

        self.workspace.create()
        try:
            self._update()
        except Exception as e:
            self._on_error(e)

    def _on_error(self, e: Exception) -> None:
        self.logger.error(f"error during update: {e}")

        # manage state
        self._on_error_handle_state()

        # manage event
        if self.runtime_cfg.on_error.action == OnErrorAction.FAIL:
            raise e

        if self.runtime_cfg.on_error.action == OnErrorAction.SKIP:
            return

        if self.runtime_cfg.on_error.action == OnErrorAction.RETRY:
            attempt = 1
            last_exception: Exception | None = e
            while attempt <= self.runtime_cfg.on_error.retry_count:
                self.logger.info(f"retrying after error: {e}")
                time.sleep(self.runtime_cfg.on_error.retry_delay)
                try:
                    self._update()
                    last_exception = None
                    break
                except Exception as ex:
                    self.logger.error(f"error during update (attempt {attempt}): {e}")
                    self._on_error_handle_state()
                    last_exception = ex
                    attempt += 1

            if last_exception:
                raise last_exception

    def _on_error_handle_state(self) -> None:
        if self.runtime_cfg.on_error.input == InputStatePolicy.DELETE:
            self.workspace.clear_input()
        if self.runtime_cfg.on_error.results == ResultStatePolicy.DELETE:
            self.workspace.clear_results()

    def __repr__(self) -> str:
        extra = []
        prefix = ""
        if getattr(self, "config", None):
            extra.append(f"config={self.config}")  # type: ignore[attr-defined]
        if extra:
            prefix = ", "
        return f"Provider(name={self.name()}, input={self.workspace.input_path}{prefix}{', '.join(extra)})"

    def results_writer(self, **kwargs: Any) -> result.Writer:
        return result.Writer(
            workspace=self.workspace,
            logger=self.logger,
            store_strategy=self.runtime_cfg.result_store,
            clear_results_before_writing=self.runtime_cfg.existing_results == ResultStatePolicy.DELETE_BEFORE_WRITE,
            **kwargs,
        )



from __future__ import annotations

import abc
import enum
import logging
import os
import time
from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, Any

import orjson
import sqlalchemy as db

if TYPE_CHECKING:
    from types import TracebackType

    from .schema import Schema
    from .workspace import Workspace


@dataclass
class Envelope:
    schema: str
    identifier: str
    item: Any


class Store:
    def __init__(
        self,
        workspace: Workspace,
        skip_duplicates: bool = False,
        logger: logging.Logger | None = None,
    ):
        self.workspace = workspace
        self.skip_duplicates = skip_duplicates
        self.start = time.time()

        if not logger:
            logger = logging.getLogger("results-writer")
        self.logger = logger

    @abc.abstractmethod
    def store(self, identifier: str, record: Envelope) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        raise NotImplementedError


class StoreStrategy(str, enum.Enum):
    FLAT_FILE = "flat-file"
    SQLITE = "sqlite"

    def __repr__(self) -> str:
        return self.value

    def store(self, *args: Any, **kwargs: Any) -> Store:
        if self == StoreStrategy.FLAT_FILE:
            return FlatFileStore(*args, **kwargs)
        if self == StoreStrategy.SQLITE:
            return SQLiteStore(*args, **kwargs)
        raise ValueError(f"unsupported result store strategy: {self!r}")


class FlatFileStore(Store):
    def store(self, identifier: str, record: Envelope) -> None:
        filename = f"{identifier}.json"
        filepath = os.path.join(self.workspace.results_path, filename)

        # allow for nested directories controlled by the provider (for large sets of results)
        if os.sep in identifier:
            os.makedirs(os.path.dirname(filepath), exist_ok=True)

        if os.path.exists(filepath):
            if self.skip_duplicates and os.path.getmtime(filepath) >= self.start:
                self.logger.warning(f"{identifier!r} entry already written (skipping)")
                return
            self.logger.trace(f"overwriting existing file: {filepath!r}")  # type: ignore[attr-defined]

        with open(filepath, "wb") as f:
            self.logger.trace(f"writing record to {filepath!r}")  # type: ignore[attr-defined]
            f.write(orjson.dumps(asdict(record), f))  # type: ignore[arg-type]

    def close(self) -> None:
        pass


class SQLiteStore(Store):
    filename = "results.db"
    table_name = "results"

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.conn = None
        self.engine = None
        self.table = None

        @db.event.listens_for(db.engine.Engine, "connect")
        def set_sqlite_pragma(dbapi_connection, connection_record):  # type: ignore[no-untyped-def]
            cursor = dbapi_connection.cursor()
            # cursor.execute("pragma journal_mode=OFF")
            cursor.execute("PRAGMA synchronous=NORMAL")
            # cursor.execute("PRAGMA cache_size=100000")
            cursor.close()

    def connection(self) -> tuple[db.engine.Connection, db.Table]:
        if not self.conn:
            self.engine = db.create_engine(f"sqlite:///{self.db_file_path}")
            self.conn = self.engine.connect()  # type: ignore[attr-defined]
            self.table = self._create_table()
        return self.conn, self.table

    @property
    def db_file_path(self) -> str:
        return os.path.join(self.workspace.results_path, self.filename)

    def _create_table(self) -> db.Table:
        metadata = db.MetaData()
        table = db.Table(
            self.table_name,
            metadata,
            db.Column("id", db.String(), primary_key=True, index=True),
            db.Column("record", db.LargeBinary()),
        )
        metadata.create_all(self.engine)
        return table

    def store(self, identifier: str, record: Envelope) -> None:
        record_str = orjson.dumps(asdict(record))
        conn, table = self.connection()

        with conn.begin():
            # upsert the record conditionally based on the skip_duplicates configuration
            existing = conn.execute(table.select().where(table.c.id == identifier)).first()
            if existing:
                if self.skip_duplicates:
                    self.logger.warning(f"{identifier!r} entry already written (skipping)")
                    return
                self.logger.trace(f"overwriting existing entry: {identifier!r}")  # type: ignore[attr-defined]
                statement = db.update(table).where(table.c.id == identifier).values(record=record_str)
            else:
                self.logger.trace(f"writing record to {identifier!r} key")  # type: ignore[attr-defined]
                statement = db.insert(table).values(id=identifier, record=record_str)

            conn.execute(statement)

    def close(self) -> None:
        if self.conn:
            self.conn.close()
            self.engine.dispose()

            self.conn = None
            self.engine = None
            self.table = None


class Writer:
    written: list[str]

    def __init__(  # noqa
        self,
        workspace: Workspace,
        logger: logging.Logger | None = None,
        skip_duplicates: bool = False,
        clear_results_before_writing: bool = False,
        store_strategy: StoreStrategy = StoreStrategy.FLAT_FILE,
    ):
        self.workspace = workspace
        self.skip_duplicates = skip_duplicates
        self.clear_results_before_writing = clear_results_before_writing

        if not logger:
            logger = logging.getLogger("results-writer")
        self.logger = logger

        self.wrote = 0
        self.store = store_strategy.store(workspace=workspace, skip_duplicates=skip_duplicates, logger=logger)

    def __enter__(self) -> Writer:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.store.close()
        self.logger.info(f"wrote {len(self)} entries")

    def __len__(self) -> int:
        return self.wrote

    def write(self, identifier: str, schema: Schema, payload: Any) -> None:
        self._clear_existing_results()

        envelope = Envelope(identifier=identifier, schema=schema.url, item=payload)

        self.store.store(identifier, envelope)

        self.wrote += 1

    def _clear_existing_results(self) -> None:
        if len(self) == 0 and self.clear_results_before_writing:
            self.workspace.clear_results()



from .generated import (
    Affected,
    Criteria,
    Criterion,
    Definition,
    Definitions,
    Evr,
    Generator,
    Metadata,
    Object,
    Objects,
    OvalDefinitions,
    Reference,
    RpminfoObject,
    RpminfoState,
    RpminfoTest,
    State,
    States,
    Tests,
)

__all__ = [
    "Affected",
    "Criteria",
    "Criterion",
    "Definition",
    "Definitions",
    "Evr",
    "Generator",
    "Metadata",
    "Object",
    "Objects",
    "OvalDefinitions",
    "Reference",
    "RpminfoObject",
    "RpminfoState",
    "RpminfoTest",
    "State",
    "States",
    "Tests",
]



from __future__ import annotations

import logging
import sys
from typing import TYPE_CHECKING, Any

from vunnel.providers import (
    alpine,
    amazon,
    chainguard,
    debian,
    github,
    mariner,
    nvd,
    oracle,
    rhel,
    sles,
    ubuntu,
    wolfi,
)

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

if TYPE_CHECKING:
    from vunnel import provider

_providers: dict[str, type[provider.Provider]] = {
    alpine.Provider.name(): alpine.Provider,
    amazon.Provider.name(): amazon.Provider,
    debian.Provider.name(): debian.Provider,
    github.Provider.name(): github.Provider,
    mariner.Provider.name(): mariner.Provider,
    nvd.Provider.name(): nvd.Provider,
    oracle.Provider.name(): oracle.Provider,
    rhel.Provider.name(): rhel.Provider,
    sles.Provider.name(): sles.Provider,
    ubuntu.Provider.name(): ubuntu.Provider,
    wolfi.Provider.name(): wolfi.Provider,
    chainguard.Provider.name(): chainguard.Provider,
}


def create(name: str, workspace_path: str, *args: Any, **kwargs: Any) -> provider.Provider:
    return _providers[name](workspace_path, *args, **kwargs)


def names() -> list[str]:
    return sorted(_providers.keys())


def versions() -> dict[str, int]:
    return {n: p.version() for (n, p) in _providers.items()}


def register(name: str, cls: type[provider.Provider]) -> None:
    if name in _providers and _providers[name] != cls:
        raise KeyError(f"provider {name!r} is already registered to another provider class: {_providers[name]!r}")
    _providers[name] = cls


def load_plugins() -> None:
    plugins = entry_points(group="vunnel.plugins.providers")

    logging.debug(f"discovered plugins: {len(plugins)}")
    for idx, p in enumerate(plugins):
        branch = "├──"
        if idx == len(plugins) - 1:
            branch = "└──"

        logging.debug(f"{branch} {p.name}: {p.value}")

    for tool in plugins:
        try:
            logging.debug(f"loading provider plugin {tool.name!r}")
            tool.load()
        except:  # noqa: E722
            # note: this should not be fatal. Log and move on.
            logging.exception(f"failed loading provider plugin {tool.name!r}")



from __future__ import annotations

import datetime  # noqa: TCH003
import json
import logging
import os
import shutil
import sqlite3
from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING, Any

import orjson
import xxhash
from dataclass_wizard import fromdict

from vunnel import schema as schemaDef
from vunnel import utils

if TYPE_CHECKING:
    from collections.abc import Generator

METADATA_FILENAME = "metadata.json"
CHECKSUM_LISTING_FILENAME = "checksums"


@dataclass
class File:
    digest: str
    path: str
    algorithm: str = "xxh64"


@dataclass
class State:
    provider: str
    urls: list[str]
    store: str
    timestamp: datetime.datetime
    version: int = 1
    listing: File | None = None
    schema: schemaDef.Schema = field(default_factory=schemaDef.ProviderStateSchema)

    @staticmethod
    def read(root: str) -> State:
        metadata_path = os.path.join(root, METADATA_FILENAME)
        with open(metadata_path, encoding="utf-8") as f:
            return fromdict(
                State,
                orjson.loads(f.read()),
            )

    def write(self, root: str, results: str, update_listing: bool = True) -> str:
        metadata_path = os.path.join(root, METADATA_FILENAME)

        if update_listing:
            listing_path = os.path.join(root, CHECKSUM_LISTING_FILENAME)
            if self.listing:
                listing_path = os.path.join(root, self.listing.path)

            if os.path.exists(listing_path):
                os.remove(listing_path)

            # why not include the file listing in the metadata file?
            # because in some cases there is a lot of data and it's easier to stream
            # the results to a tab-delimited file than include in the metadata file
            self.listing = File(
                digest=write_file_listing(listing_path, results),
                algorithm="xxh64",
                path=os.path.basename(listing_path),  # may have been overridden, keep value
            )

        with open(metadata_path, "w") as f:
            json.dump(asdict(self), f, indent=2, cls=utils.DTEncoder)

        return metadata_path

    def result_files(self, root: str) -> Generator[File, File, None]:
        if self.listing:
            full_path = os.path.join(root, self.listing.path)
            with open(full_path) as f:
                for digest, filepath in (line.split() for line in f.readlines()):
                    yield File(digest=digest, path=filepath, algorithm=self.listing.algorithm)

    def result_count(self, root: str) -> int:
        count = 0
        if self.listing:
            full_path = os.path.join(root, self.listing.path)
            with open(full_path) as f:
                for _digest, filepath in (line.split() for line in f.readlines()):
                    if filepath.endswith(".db"):
                        # open up the sqlite db and count the records in the "results" table
                        with sqlite3.connect(os.path.join(root, filepath)) as db:
                            count += db.execute("SELECT COUNT(*) FROM results").fetchone()[0]
                    else:
                        count += 1

        return count


class Workspace:
    def __init__(self, root: str, name: str, create: bool = False, logger: logging.Logger | None = None):
        if not logger:
            logger = logging.getLogger(self.__class__.__name__)
        self.logger = logger
        self._root = root
        self.name = name

        if create:
            self.create()

    @property
    def path(self) -> str:
        return os.path.join(self._root, self.name)

    @property
    def results_path(self) -> str:
        return os.path.join(self.path, "results")

    @property
    def input_path(self) -> str:
        return os.path.join(self.path, "input")

    def create(self) -> None:
        if not os.path.exists(self.input_path):
            self.logger.debug(f"creating input workspace {self.input_path!r}")
            os.makedirs(self.input_path)
        else:
            self.logger.debug(f"using existing input workspace {self.input_path!r}")

        if not os.path.exists(self.results_path):
            self.logger.debug(f"creating results workspace {self.results_path!r}")
            os.makedirs(self.results_path)
        else:
            self.logger.debug(f"using existing results workspace {self.results_path!r}")

    def clear(self) -> None:
        self.clear_input()
        self.clear_results()
        self._clear_metadata()

    def _clear_metadata(self) -> None:
        utils.silent_remove(os.path.join(self.path, METADATA_FILENAME))
        utils.silent_remove(os.path.join(self.path, CHECKSUM_LISTING_FILENAME))

    def clear_results(self) -> None:
        if os.path.exists(self.results_path):
            self.logger.debug("clearing existing results")
            shutil.rmtree(self.results_path)
            os.makedirs(self.results_path, exist_ok=True)

        try:
            current_state = State.read(root=self.path)
            current_listing = None
            if current_state.listing:
                current_listing = current_state.listing.path
            current_state.listing = None
            current_state.write(self.path, self.results_path)
            if current_listing:
                utils.silent_remove(current_listing)
        except FileNotFoundError:
            pass

    def clear_input(self) -> None:
        if os.path.exists(self.input_path):
            self.logger.debug("clearing existing input")
            shutil.rmtree(self.input_path)
            os.makedirs(self.input_path, exist_ok=True)

    def record_state(self, version: int, timestamp: datetime.datetime, urls: list[str], store: str) -> None:
        try:
            current_state = State.read(root=self.path)
        except FileNotFoundError:
            current_state = None

        if current_state:
            if not urls:
                urls = current_state.urls
            if not timestamp:
                timestamp = current_state.timestamp

        self.logger.info("recording workspace state")

        state = State(provider=self.name, version=version, urls=urls, store=store, timestamp=timestamp)
        metadata_path = state.write(self.path, self.results_path)

        self.logger.debug(f"wrote workspace state to {metadata_path}")

    def state(self) -> State | None:
        return State.read(self.path)


def digest_path_with_hasher(path: str, hasher: Any, label: str | None, size: int = 65536) -> str:
    with open(path, "rb") as f:
        while b := f.read(size):
            hasher.update(b)

    if label:
        return label + ":" + hasher.hexdigest()
    return hasher.hexdigest()


# def sha256_digest(path: str, label: bool = True) -> str:
#     return digest_path_with_hasher(path, hashlib.sha256(), "sha256" if label else None)


def xxhash64_digest(path: str, label: bool = True) -> str:
    return digest_path_with_hasher(path, xxhash.xxh64(), "xxh64" if label else None)


def write_file_listing(output_file: str, path: str) -> str:
    listing_hasher = xxhash.xxh64()

    with open(output_file, "w", encoding="utf-8") as f:
        for root, dirs, files in os.walk(path):
            dirs.sort()  # sort the existing list that os.walk generator continues to reference
            for file in sorted(files):
                full_path = os.path.join(root, file)
                path_relative_to_results = os.path.relpath(full_path, path)
                path_relative_to_workspace = os.path.join(os.path.basename(path), path_relative_to_results)

                contents = f"{xxhash64_digest(full_path, label=False)}  {path_relative_to_workspace}\n"
                listing_hasher.update(contents.encode("utf-8"))

                f.write(contents)

    return listing_hasher.hexdigest()



from __future__ import annotations

import copy
import os
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from vunnel import provider, result, schema

from .parser import Parser

if TYPE_CHECKING:
    import datetime


@dataclass
class Config:
    token: str = "env:GITHUB_TOKEN"
    api_url: str = "https://api.github.com/graphql"
    runtime: provider.RuntimeConfig = field(
        default_factory=lambda: provider.RuntimeConfig(
            result_store=result.StoreStrategy.SQLITE,
            existing_results=provider.ResultStatePolicy.DELETE_BEFORE_WRITE,
        ),
    )
    request_timeout: int = 125

    def __post_init__(self) -> None:
        if self.token.startswith("env:"):
            self.token = os.environ.get(self.token[4:], "")

    def __str__(self) -> str:
        # sanitize secrets from any output
        tok_value = self.token
        str_value = super().__str__()
        if not tok_value:
            return str_value
        return str_value.replace(tok_value, "********")


class Provider(provider.Provider):
    def __init__(self, root: str, config: Config | None = None):
        if not config:
            config = Config()
        super().__init__(root, runtime_cfg=config.runtime)
        self.config = config

        self.logger.debug(f"config: {config}")

        self.schema = schema.GithubSecurityAdvisorySchema()
        self.parser = Parser(
            workspace=self.workspace,
            token=config.token,
            api_url=config.api_url,
            download_timeout=config.request_timeout,
            logger=self.logger,
        )

    @classmethod
    def name(cls) -> str:
        return "github"

    def update(self, last_updated: datetime.datetime | None) -> tuple[list[str], int]:
        namespace = "github"
        with self.results_writer() as writer:
            for advisory in self.parser.get():
                all_fixes = copy.deepcopy(advisory.get("FixedIn")) if isinstance(advisory.get("FixedIn"), list) else []
                for ecosystem in advisory.ecosystems:
                    advisory["namespace"] = f"{namespace}:{ecosystem}"

                    # filter the list of fixes for this ecosystem
                    advisory["FixedIn"] = [item for item in all_fixes if item.get("ecosystem") == ecosystem]

                    vuln_id = advisory["ghsaId"]

                    namespace = namespace.lower()
                    ecosystem = ecosystem.lower()
                    vuln_id = vuln_id.lower()

                    writer.write(
                        identifier=os.path.join(f"{namespace}:{ecosystem}", vuln_id),
                        schema=self.schema,
                        payload={"Vulnerability": {}, "Advisory": dict(advisory)},
                    )

        return [self.parser.api_url], len(writer)



"""
RPM utilities with no binary dependencies on rpm or rpmUtil.
"""
from __future__ import annotations


def parse_version(rpm_version: str) -> list[str]:
    """
    Return a tuple from the given version string.

    :param rpm_version:
    :return:
    """
    return rpm_version.split("-")


def split_rpm_filename(rpm_filename: str) -> tuple[str | None, str, str, str | None, str]:
    """
    Parse the components of an rpm filename and return them as a tuple: (name, version, release, epoch, arch)

    foo-1.0-1.x86_64.rpm -> foo, 1.0, 1, '', x86_64
    1:bar-9-123a.ia64.rpm -> bar, 9, 123a, 1, ia64

    :param rpm_filename: a string filename (not path) of an rpm file
    :returns: a tuple of the constituent parts compliant with RPM spec.
    """

    components = rpm_filename.rsplit(".rpm", 1)[0].rsplit(".", 1)
    arch = components.pop()

    rel_comp = components[0].rsplit("-", 2)
    release = rel_comp.pop()

    # Version
    version = rel_comp.pop()

    # Epoch
    epoch_comp = rel_comp[0].split(":", 1) if rel_comp else []
    if len(epoch_comp) == 1:
        epoch = ""
        name = epoch_comp[0]
    elif len(epoch_comp) > 1:
        epoch = epoch_comp[0]
        name = epoch_comp[1]
    else:
        epoch = None
        name = None

    return name, version, release, epoch, arch


def split_fullversion(version: str) -> tuple[str | None, str | None, str | None]:
    """
    Splits version string into a tuple (epoch, version, release). Inbound version string is specific to anchore engine's
    implementation of versions and may not compliant with the rpm spec.

    Use this function for splitting versions already processed by anchore engine.
    For parsing info from rpm file name, use split_rpm_filename()

    '2.27-34.base.el7' -> (null, '2.27', '34.base.el7')
    '1:2.27-34.base.el7' -> ('1', '2.27', '34.base.el7')

    :param version: Version string with or without an epoch prefix
    :return: tuple (epoch, version, release). epoch is null if the inbound version does not contain epoch
    """
    ver_comp = version.rsplit("-", 1)

    release = ver_comp.pop() if len(ver_comp) > 1 else None

    epoch_comp = ver_comp[0].split(":", 1) if ver_comp else []

    if len(epoch_comp) == 1:
        epoch = None
        discovered_version = epoch_comp[0]
    elif len(epoch_comp) > 1:
        epoch = epoch_comp[0]
        discovered_version = epoch_comp[1]
    else:
        epoch = None
        discovered_version = None

    return (
        epoch,
        discovered_version,
        release,
    )


def compare_versions(ver_a: str, ver_b: str) -> int:
    """
    Compare pkg and versions using anchore engine rules.
    Follows standard __cmp__ semantics of -1 iff a < b, 0 iff a == b, 1 iff a > b

    :param ver_a:
    :param ver_b:
    :return:
    """
    if ver_a == ver_b:
        return 0

    (e1, v1, r1) = split_fullversion(ver_a)
    (e2, v2, r2) = split_fullversion(ver_b)

    return compare_labels((e1, v1, r1), (e2, v2, r2))


def compare_labels(evr_1: tuple[str | None, str | None, str | None], evr_2: tuple[str | None, str | None, str | None]) -> int:
    """
    Compare the EVR label tuples (epoch, version, release).

    :param evr_1:
    :param evr_2:
    :return: -1, 0, 1 is standard __cmp__ semantics
    """
    epoch_1, ver_1, rel_1 = evr_1
    epoch_2, ver_2, rel_2 = evr_2

    if epoch_1 is not None and epoch_2 is not None:  # compare only when both epochs are available. ignore otherwise
        if epoch_1 > epoch_2:
            return 1
        if epoch_1 < epoch_2:
            return -1

    cmp_result = rpm_ver_cmp(ver_1, ver_2)

    if cmp_result != 0:
        return cmp_result

    return rpm_ver_cmp(rel_1, rel_2)


def rpm_ver_cmp(a: str | None, b: str | None) -> int:  # noqa: C901,PLR0912,PLR0911
    """
    A translation of the RPM lib's C code for version compare rpmvercmp in lib/rpmvercmp.c into pure python with
    no external

    compare alpha and numeric segments of two versions
    return 1: a is newer than b
        0: a and b are the same version
       -1: b is newer than a
    """
    if a is None:
        a = ""
    if b is None:
        b = ""

    # Convert to a list of single chars
    l_a = list(a.strip())
    l_b = list(b.strip())
    is_num = False

    # loop through each version segment of str1 and str2 and compare them
    while l_a and l_b:
        # Skip any whitespace preceding
        while l_a and not l_a[0].isalnum():
            l_a.pop(0)
        while l_b and not l_b[0].isalnum():
            l_b.pop(0)

        # If we ran to the end of either, we are finished with the loop
        if not (l_a and l_b):
            break

        # Get the next numeric or alpha segment to compare. Must get similar types from both strings.
        # This is a greedy consumption from the src string. so l_a/l_b will be truncated from the front to
        # construct the a/b_seg
        is_num, a_seg = greedy_find_block(l_a)
        b_is_num, b_seg = greedy_find_block(l_b, is_num)

        # this cannot happen, as we previously tested to make sure that
        # the first string has a non-null segment
        # if (one == str1) return -1;     # arbitrary
        if l_a == a_seg:
            raise Exception("Encountered null segment in str. Unexpected")
            # return -1 # Arbitrary per C impl

        # take care of the case where the two version segments are
        # different types: one numeric, the other alpha (i.e. empty)
        # numeric segments are always newer than alpha segments
        # XXX See patch #60884 (and details) from bugzilla #50977.
        # if (two == str2) return isnum ? 1 : -1;
        if l_b == b_seg:
            return 1 if is_num else -1

        a_seg_str = "".join(a_seg)
        b_seg_str = "".join(b_seg)

        if is_num:
            # Strip leading zeros since this is a numeric comparison
            a_seg_str = a_seg_str.lstrip("0")
            b_seg_str = b_seg_str.lstrip("0")

            # whichever number has more digits wins
            if len(a_seg_str) > len(b_seg_str):
                return 1
            if len(a_seg_str) < len(b_seg_str):
                return -1

        # String compare of the segments, covers both numeric and non since they are same length if numeric
        if a_seg_str > b_seg_str:
            return 1
        if a_seg_str < b_seg_str:
            return -1

    # this catches the case where all numeric and alpha segments have
    # compared identically but the segment separating characters were
    # different

    # zhill - note the original comparison: if ((!*one) && (!*two)) return 0;
    if not l_a and not l_b:
        return 0

    # whichever version still has characters left over wins
    if not l_a:
        return -1
    return 1


def greedy_find_block(list_str: list[str], expected_digit: bool | None = None) -> tuple[bool, list[str]]:
    """
    Scan the string and return the substring, index, and type of the next block.
    A block is defined as a contiguous set of numeric or alpha characters. The point at which the string
    converts from one to another is the edge of the block.

    Will pop elements of the list to consume them during processing

    :param list_str: a string in list form: ['a', 'c', 'd', '1', '0', '.']
    :return: (bool, list) tuple, where bool is isdigit() for first char of string
    """

    # True for digits, false for alpha
    chr_type = list_str[0].isdigit()
    if expected_digit is not None and expected_digit != chr_type:
        # An explicit type request and the head of this string doesn't match, so return the other type and an empty list
        return expected_digit, []

    result: list[str] = []
    while list_str and chr_type == list_str[0].isdigit():
        result += list_str.pop(0)

    return chr_type, result



# flake8: noqa
from __future__ import annotations

import datetime
import logging
import urllib.parse
from typing import Any, Generator

import orjson
import requests

from vunnel import utils


class NvdAPI:
    _cve_api_url_: str = "https://services.nvd.nist.gov/rest/json/cves/2.0"
    _cve_history_api_url_: str = "https://services.nvd.nist.gov/rest/json/cvehistory/2.0"
    _max_results_per_page_: int = 2000
    max_date_range_days: int = 120

    def __init__(self, api_key: str | None = None, logger: logging.Logger | None = None, timeout: int = 30):
        self.api_key = api_key
        self.timeout = timeout

        if not logger:
            logger = logging.getLogger(self.__class__.__name__)
        self.logger = logger

    def cve_history(
        self,
        cve_id: str | None = None,
        results_per_page: int
        | None = None,  # from api docs: "it is recommended that users of the CVE API use the default resultsPerPage value"
        change_start_date: str
        | datetime.datetime
        | None = None,  # note: if you specify a changeStartDate, you must also specify a changeEndDate
        change_end_date: str | datetime.datetime | None = None,  # note: maximum date range is 120 days
    ) -> Generator[dict[str, Any], Any, None]:
        parameters = {}

        if cve_id:
            parameters["cveId"] = cve_id

        if results_per_page:
            if results_per_page < 1 or results_per_page > self._max_results_per_page_:
                raise RuntimeError(f"'resultsPerPage' must be > 1 and < {self._max_results_per_page_}")
            parameters["resultsPerPage"] = str(results_per_page)

        if change_start_date:
            parameters["changeStartDate"] = clean_date(change_start_date)

        if change_end_date:
            parameters["changeEndDate"] = clean_date(change_end_date)

        return self._request_all_pages(
            self._cve_history_api_url_,
            parameters,
            message="fetching CVE history",
        )

    def cve(
        self,
        cve_id: str | None = None,
        results_per_page: int
        | None = None,  # from api docs: "it is recommended that users of the CVE API use the default resultsPerPage value"
        last_mod_start_date: str | datetime.datetime | None = None,
        last_mod_end_date: str | datetime.datetime | None = None,
        pub_start_date: str
        | datetime.datetime
        | None = None,  # note: if you specify a pubStartDate, you must also specify a pubEndDate
        pub_end_date: str | datetime.datetime | None = None,  # note: maximum date range is 120 days
    ) -> Generator[dict[str, Any], Any, None]:
        parameters = {}

        if cve_id:
            parameters["cveId"] = cve_id

        if results_per_page:
            if results_per_page < 1 or results_per_page > self._max_results_per_page_:
                raise RuntimeError(f"'resultsPerPage' must be > 1 and < {self._max_results_per_page_}")
            parameters["resultsPerPage"] = str(results_per_page)

        if last_mod_start_date:
            parameters["lastModStartDate"] = clean_date(last_mod_start_date)

        if last_mod_end_date:
            parameters["lastModEndDate"] = clean_date(last_mod_end_date)

        if pub_start_date:
            parameters["pubStartDate"] = clean_date(pub_start_date)

        if pub_end_date:
            parameters["pubEndDate"] = clean_date(pub_end_date)

        yield from self._request_all_pages(
            self._cve_api_url_,
            parameters,
            message="fetching CVE results",
        )

    def _request_all_pages(
        self, url: str, parameters: dict[str, str], message: str = "fetching results"
    ) -> Generator[dict[str, Any], Any, None]:
        headers = {
            "content-type": "application/json",
        }

        if self.api_key:
            headers["apiKey"] = self.api_key

        self.logger.debug(f"{message}")

        response = self._request(url, parameters, headers)

        payload = orjson.loads(response.text)
        if "message" in payload:
            raise RuntimeError(f"API error: {response['message']}")

        yield payload

        total_results = payload["totalResults"]
        results_per_page = payload["resultsPerPage"]

        if not total_results or not results_per_page:
            return

        pages = total_results // results_per_page
        index = results_per_page

        for page in range(pages):
            self.logger.debug(f"{message} (page {page+2} of {pages})")

            parameters["resultsPerPage"] = str(results_per_page)
            parameters["startIndex"] = str(index)

            response = self._request(self._cve_api_url_, parameters, headers)

            yield orjson.loads(response.text)

            index += results_per_page

    @utils.retry_with_backoff()
    def _request(self, url: str, parameters: dict[str, str], headers: dict[str, str]) -> requests.Response:
        # this is to prevent from encoding the ':' in any timestamps passed
        # (e.g. prevent pubStartDate=2002-01-01T00%3A00%3A00 , want pubStartDate=2002-01-01T00:00:00)
        payload_str = urllib.parse.urlencode(parameters, safe=":")

        response = requests.get(url, params=payload_str, headers=headers, timeout=self.timeout)
        response.encoding = "utf-8"
        response.raise_for_status()

        return response


def clean_date(dt: datetime.datetime | str) -> str:
    if isinstance(dt, datetime.datetime):
        return dt.isoformat()
    return datetime.datetime.strptime(dt, "%Y-%m-%d %H:%M").isoformat()



# flake8: noqa
from __future__ import annotations

import datetime
import logging
import os
from typing import Any, Generator

from vunnel.workspace import Workspace

from .api import NvdAPI


class Manager:
    def __init__(
        self,
        workspace: Workspace,
        logger: logging.Logger | None = None,
        download_timeout: int = 125,
        api_key: str | None = None,
    ) -> None:
        self.workspace = workspace

        if not logger:
            logger = logging.getLogger(self.__class__.__name__)
        self.logger = logger

        self.api = NvdAPI(api_key=api_key, logger=logger, timeout=download_timeout)
        self.urls = [self.api._cve_api_url_]

    def get(
        self, last_updated: datetime.datetime | None, skip_if_exists: bool = False
    ) -> Generator[tuple[str, dict[str, Any]], Any, None]:
        if skip_if_exists and self._can_update_incrementally(last_updated):
            yield from self._download_updates(last_updated)  # type: ignore
        else:
            yield from self._download_all()

    def _can_update_incrementally(self, last_updated: datetime.datetime | None) -> bool:
        if not last_updated:
            return False

        now = datetime.datetime.now(tz=datetime.timezone.utc)
        days_since_last_sync = (now - last_updated).days

        if days_since_last_sync >= NvdAPI.max_date_range_days:
            self.logger.info(
                f"last sync was {days_since_last_sync} days ago (more than {NvdAPI.max_date_range_days} days, the max range value of the NVD API), downloading all data"
            )
            return False

        return True

    def _download_all(self) -> Generator[tuple[str, dict[str, Any]], Any, None]:
        self.logger.info("downloading all CVEs")

        # TODO: should we delete all existing state in this case first?

        for response in self.api.cve():
            yield from self._unwrap_records(response)

    def _download_updates(self, last_updated: datetime.datetime) -> Generator[tuple[str, dict[str, Any]], Any, None]:
        self.logger.debug(f"downloading CVEs changed since {last_updated.isoformat()}")

        # get the list of CVEs that have been updated since the last sync
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        for idx, response in enumerate(self.api.cve(last_mod_start_date=last_updated, last_mod_end_date=now)):
            if idx == 0:
                total_results = response.get("totalResults", None)
                if total_results:
                    self.logger.debug(f"discovered {total_results} updated CVEs")

            yield from self._unwrap_records(response)

    def _unwrap_records(self, response: dict[str, Any]) -> Generator[tuple[str, dict[str, Any]], Any, None]:
        for vuln in response["vulnerabilities"]:
            cve_id = vuln["cve"]["id"]
            year = cve_id.split("-")[1]
            yield os.path.join(year, cve_id), vuln



import os
import tempfile
from subprocess import PIPE, Popen

import requests

MARINER_URL_BASE = "https://raw.githubusercontent.com/microsoft/CBL-MarinerVulnerabilityData/main/{}"
MARINER_URL_FILENAME = "cbl-mariner-{}-oval.xml"


def download_version(version: str, dest_dir: str) -> None:
    filename = MARINER_URL_FILENAME.format(version)
    url = MARINER_URL_BASE.format(filename)
    r = requests.get(url, timeout=125)
    destination = os.path.join(dest_dir, filename)
    with open(destination, "wb") as w:
        w.write(r.content)


def main() -> None:
    versions = ["2.0"]
    dest_path = tempfile.TemporaryDirectory()
    for v in versions:
        download_version(v, dest_path.name)

    script_dir = os.path.realpath(os.path.dirname(__file__))
    args = [
        "xsdata",
        "generate",
        dest_path.name,
        "-r",
        "--relative-imports",
        "--compound-fields",
        "--package",
        "model.generated",
        "--structure-style",
        "single-package",
    ]
    process = Popen(args=args, stderr=PIPE, stdout=PIPE, cwd=script_dir)
    stdout, stderr = process.communicate()
    print(stdout)
    print(stderr)
    pass


if __name__ == "__main__":
    main()
